import os, stat, shutil

def remove_dir_with_permissions(dir_path):
    def handle_permission_error(func, path, exc_info):
        os.chmod(path, stat.S_IWUSR)
        func(path)

    if os.path.exists(dir_path):
        shutil.rmtree(dir_path, onerror=handle_permission_error)


from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.io.ase import AseAtomsAdaptor
from ase.io import read
from pymatgen.core.structure import Structure

import os,juliacall
from CoREMOF.calculation.mof_collection import MofCollection
from CoREMOF.utils import remove

from molSimplify.Informatics.MOF.MOF_descriptors import get_MOF_descriptors


def SpaceGroup(structure):
    result_sg = {}
    result_sg["unit"]="nan"
    atoms = read(structure)
    structure_ = AseAtomsAdaptor.get_structure(atoms)
    result_ = SpacegroupAnalyzer(structure_, symprec=0.01, angle_tolerance=5)
    hall_symbol = result_.get_hall()
    space_group_number = result_.get_space_group_number()
    crystal_system = result_.get_crystal_system()
    result_sg["hall_symbol"]=hall_symbol
    result_sg["space_group_number"]=space_group_number
    result_sg["crystal_system"]=crystal_system

    return result_sg

def Mass(structure):
    result_m = {}
    result_m["unit"]="amu"
    atoms = read(structure)
    total_mass = atoms.get_masses().sum()
    result_m["total_mass"]=float(total_mass)

    return result_m

def Volume(structure):
    result_v = {}
    result_v["unit"]="Å^3"
    with open(structure, "r", encoding="utf-8") as f:
        cif_file = f.read()
    atoms = Structure.from_str(cif_file, fmt="cif")
    total_volume = atoms.volume
    result_v["total_volume"]=total_volume

    return result_v

def n_atom(structure):
    result_na = {}
    result_na["unit"]="nan"
    atoms = read(structure)
    number_atoms = len(atoms)
    result_na["number_atoms"]=number_atoms

    return result_na

def topology(structure,node_type="single"):

    package_directory = os.path.abspath(__file__).replace("mof_features.py","")
    os.environ["JULIA_DEPOT_PATH"] = package_directory

    juliacall.Main.seval('import Pkg; Pkg.add("CrystalNets")')

    jl = juliacall.newmodule("topo")
    jl.seval("using CrystalNets")

    options = jl.CrystalNets.Options(structure=jl.StructureType.MOF)
    result = jl.determine_topology(structure, options)
    result_tp = {}
    result_save = []
    for x in result:
        if node_type=="single":
            result_ = x[0][jl.Clustering.SingleNodes]
            result_save.append(result_)
        elif node_type=="all":
            result_ = x[0][jl.Clustering.AllNodes]
            result_save.append(result_)

    result_tp["dimention"]=[]
    result_tp["topology"] = []
    for info in result_save:
        
        result_tp["dimention"].append(jl.ndims(info.genome))
        result_tp["topology"].append(str(info))

    return result_tp


def get_oms_file(structure):
    a_mof_collection = MofCollection(path_list = [structure], 
                                 analysis_folder="tmp_oms")
    a_mof_collection.analyse_mofs(num_batches=1,overwrite=False)
    oms_result = {
                    "Metal Types": a_mof_collection.mof_oms_df["Metal Types"][structure.replace(".cif","").split("/")[-1]],
                    "Has OMS": a_mof_collection.mof_oms_df["Has OMS"][structure.replace(".cif","").split("/")[-1]],
                    "OMS Types": a_mof_collection.mof_oms_df["OMS Types"][structure.replace(".cif","").split("/")[-1]],
                }

    remove.remove_dir_with_permissions("tmp_oms")

    return oms_result


def get_oms_folder(input_folder, n_batch = 1):
    mof_collection = MofCollection.from_folder(collection_folder = input_folder, 
                                                analysis_folder="tmp_oms")
    mof_collection.analyse_mofs(num_batches=n_batch,overwrite=False)
    oms_result = {}
    for name, row in mof_collection.mof_oms_df.iterrows():
        
        oms_result[name] = {
                            "Metal Types": row[0],
                            "Has OMS": row[1],
                            "OMS Types": row[2]
                        }

    remove.remove_dir_with_permissions("tmp_oms")

    return oms_result


def RACs(structure):

    metal_fnames = ['D_mc-I-0-all', 'D_mc-I-1-all', 'D_mc-I-2-all', 'D_mc-I-3-all', 'D_mc-S-0-all', 'D_mc-S-1-all', 'D_mc-S-2-all',
    'D_mc-S-3-all', 'D_mc-T-0-all', 'D_mc-T-1-all', 'D_mc-T-2-all', 'D_mc-T-3-all', 'D_mc-Z-0-all', 'D_mc-Z-1-all', 'D_mc-Z-2-all',
    'D_mc-Z-3-all', 'D_mc-chi-0-all', 'D_mc-chi-1-all', 'D_mc-chi-2-all', 'D_mc-chi-3-all', 'f-I-0-all', 'f-I-1-all', 'f-I-2-all',
    'f-I-3-all', 'f-S-0-all', 'f-S-1-all', 'f-S-2-all', 'f-S-3-all', 'f-T-0-all', 'f-T-1-all', 'f-T-2-all', 'f-T-3-all', 'f-Z-0-all',
    'f-Z-1-all', 'f-Z-2-all', 'f-Z-3-all', 'f-chi-0-all', 'f-chi-1-all', 'f-chi-2-all', 'f-chi-3-all', 'mc-I-0-all', 'mc-I-1-all',
    'mc-I-2-all', 'mc-I-3-all', 'mc-S-0-all', 'mc-S-1-all', 'mc-S-2-all', 'mc-S-3-all', 'mc-T-0-all', 'mc-T-1-all', 'mc-T-2-all',
    'mc-T-3-all', 'mc-Z-0-all', 'mc-Z-1-all', 'mc-Z-2-all', 'mc-Z-3-all', 'mc-chi-0-all', 'mc-chi-1-all', 'mc-chi-2-all', 'mc-chi-3-all']

    linker_fnames = ['D_lc-I-0-all', 'D_lc-I-1-all', 'D_lc-I-2-all', 'D_lc-I-3-all', 'D_lc-S-0-all', 'D_lc-S-1-all', 'D_lc-S-2-all',
    'D_lc-S-3-all', 'D_lc-T-0-all', 'D_lc-T-1-all', 'D_lc-T-2-all', 'D_lc-T-3-all', 'D_lc-Z-0-all', 'D_lc-Z-1-all', 'D_lc-Z-2-all',
    'D_lc-Z-3-all', 'D_lc-alpha-0-all', 'D_lc-alpha-1-all', 'D_lc-alpha-2-all', 'D_lc-alpha-3-all', 'D_lc-chi-0-all', 'D_lc-chi-1-all',
    'D_lc-chi-2-all', 'D_lc-chi-3-all', 'lc-I-0-all', 'lc-I-1-all', 'lc-I-2-all', 'lc-I-3-all', 'lc-S-0-all', 'lc-S-1-all', 'lc-S-2-all',
    'lc-S-3-all', 'lc-T-0-all', 'lc-T-1-all', 'lc-T-2-all', 'lc-T-3-all', 'lc-Z-0-all', 'lc-Z-1-all', 'lc-Z-2-all', 'lc-Z-3-all',
    'lc-alpha-0-all', 'lc-alpha-1-all', 'lc-alpha-2-all', 'lc-alpha-3-all', 'lc-chi-0-all', 'lc-chi-1-all', 'lc-chi-2-all', 'lc-chi-3-all',
    'f-lig-I-0', 'f-lig-I-1', 'f-lig-I-2', 'f-lig-I-3', 'f-lig-S-0', 'f-lig-S-1', 'f-lig-S-2', 'f-lig-S-3', 'f-lig-T-0', 'f-lig-T-1',
    'f-lig-T-2', 'f-lig-T-3', 'f-lig-Z-0', 'f-lig-Z-1', 'f-lig-Z-2', 'f-lig-Z-3', 'f-lig-chi-0', 'f-lig-chi-1', 'f-lig-chi-2', 'f-lig-chi-3']

    fg_fnames = ['D_func-I-0-all', 'D_func-I-1-all', 'D_func-I-2-all', 'D_func-I-3-all', 'D_func-S-0-all', 'D_func-S-1-all',
    'D_func-S-2-all', 'D_func-S-3-all', 'D_func-T-0-all', 'D_func-T-1-all', 'D_func-T-2-all', 'D_func-T-3-all', 'D_func-Z-0-all',
    'D_func-Z-1-all', 'D_func-Z-2-all', 'D_func-Z-3-all', 'D_func-alpha-0-all', 'D_func-alpha-1-all', 'D_func-alpha-2-all',
    'D_func-alpha-3-all', 'D_func-chi-0-all', 'D_func-chi-1-all', 'D_func-chi-2-all', 'D_func-chi-3-all', 'func-I-0-all',
    'func-I-1-all', 'func-I-2-all', 'func-I-3-all', 'func-S-0-all', 'func-S-1-all', 'func-S-2-all', 'func-S-3-all', 'func-T-0-all',
    'func-T-1-all', 'func-T-2-all', 'func-T-3-all', 'func-Z-0-all', 'func-Z-1-all', 'func-Z-2-all', 'func-Z-3-all', 'func-alpha-0-all',
    'func-alpha-1-all', 'func-alpha-2-all', 'func-alpha-3-all', 'func-chi-0-all', 'func-chi-1-all', 'func-chi-2-all', 'func-chi-3-all']

    result_rac = {}
    result_rac["Metal"] = {}
    result_rac["Linker"] = {}
    result_rac["Function-group"] = {}

    os.makedirs("tmp_rac", exist_ok=True)
    
    name = os.path.basename(structure).replace(".cif", "")
    
    full_names, full_descriptors = get_MOF_descriptors(
                                                        structure,
                                                        3,
                                                        path='tmp_rac',
                                                        xyzpath=f'tmp_rac/{name}.xyz',
                                                        max_num_atoms=6000
                                                        )
                                                        
    descriptor_data = dict(zip(full_names, full_descriptors))

    for metal in metal_fnames:
        result_rac["Metal"][metal] =  round(float(descriptor_data[metal]), 4)
    for linker in linker_fnames:
        result_rac["Linker"][linker] =  round(float(descriptor_data[linker]), 4)
    for fg in fg_fnames:
        result_rac["Function-group"][fg] =  round(float(descriptor_data[fg]), 4)

    remove.remove_dir_with_permissions("tmp_rac")

    return result_rac



"""
rare-earth metal and abundance got from wiki (https://en.wikipedia.org/wiki/Rare-earth_element)
"""

EXPENSIVE_METALS = {
                    "Au",  # Gold
                    "Ag",  # Silver
                    "Dy",  # Dysprosium
                    "Eu",  # Europium
                    "Ga",  # Gallium
                    "Gd",  # Gadolinium
                    "Hf",  # Hafnium
                    "In",  # Indium
                    "Ir",  # Iridium
                    "La",  # Lanthanum
                    "Mo",  # Molybdenum
                    "Nd",  # Neodymium
                    "Pd",  # Palladium
                    "Pr",  # Praseodymium
                    "Pt",  # Platinum
                    "Rh",  # Rhodium
                    "Ru",  # Ruthenium
                    "Se",  # Selenium
                    "Sm",  # Samarium
                    "Tb",  # Terbium
                    "Te",  # Tellurium
                    "Tm",  # Thulium
                    "U",   # Uranium
                    "Y"    # Yttrium
                }

RARE_METALS = {
                "Sc": 22,  # Scandium, 22 ppm
                "Y": 33,   # Yttrium, 33 ppm
                "La": 39,  # Lanthanum, 39 ppm
                "Ce": 66.5,  # Cerium, 66.5 ppm
                "Pr": 9.2,  # Praseodymium, 9.2 ppm
                "Nd": 41.5,  # Neodymium, 41.5 ppm
                "Pm": 10e-15,  # Promethium, 10e-15 ppm
                "Sm": 7.05,  # Samarium, 7.05 ppm
                "Eu": 2,  # Europium, 2 ppm
                "Gd": 6.2,  # Gadolinium, 6.2 ppm
                "Tb": 1.2,  # Terbium, 1.2 ppm
                "Dy": 5.2,  # Dysprosium, 5.2 ppm
                "Ho": 1.3,  # Holmium, 1.3 ppm
                "Er": 3.5,  # Erbium, 3.5 ppm
                "Tm": 0.52,  # Thulium, 0.52 ppm
                "Yb": 3.2,  # Ytterbium, 3.2 ppm
                "Lu":0.8   # Lutetium, 0.8 ppm
            }



cv_features_minimal = ['site Number',
        'site AtomicWeight',
        'site Row',
        'site Column',
        'site Electronegativity',
        'site CovalentRadius',
        'AGNI eta=8.00e-01',
        'AGNI eta=1.23e+00',
        'AGNI eta=1.88e+00',
        'AGNI eta=2.89e+00',
        'AGNI eta=4.43e+00',
        'AGNI eta=6.80e+00',
        'AGNI eta=1.04e+01',
        'AGNI eta=1.60e+01',
        'G2_0.05',
        'G2_4.0',
        'G2_20.0',
        'G2_80.0',
        'G4_0.005_1.0_1.0',
        'G4_0.005_1.0_-1.0',
        'G4_0.005_4.0_1.0',
        'G4_0.005_4.0_-1.0',
        ]
cv_features= ['site Number',
        'site AtomicWeight',
        'site Row',
        'site Column',
        'site Electronegativity',
        'site CovalentRadius',
        'AGNI eta=8.00e-01',
        'AGNI eta=1.23e+00',
        'AGNI eta=1.88e+00',
        'AGNI eta=2.89e+00',
        'AGNI eta=4.43e+00',
        'AGNI eta=6.80e+00',
        'AGNI eta=1.04e+01',
        'AGNI eta=1.60e+01',
        'G2_0.05',
        'G2_4.0',
        'G2_20.0',
        'G2_80.0',
        'G4_0.005_1.0_1.0',
        'G4_0.005_1.0_-1.0',
        'G4_0.005_4.0_1.0',
        'G4_0.005_4.0_-1.0',
        'local difference in Number',
        'local difference in AtomicWeight',
        'local difference in Row',
        'local difference in Column',
        'local difference in Electronegativity',
        'local difference in CovalentRadius',
        'local signed difference in Number',
        'local signed difference in AtomicWeight',
        'local signed difference in Row',
        'local signed difference in Column',
        'local signed difference in Electronegativity',
        'local signed difference in CovalentRadius',
        'maximum local difference in Number',
        'maximum local difference in AtomicWeight',
        'maximum local difference in Row',
        'maximum local difference in Column',
        'maximum local difference in Electronegativity',
        'maximum local difference in CovalentRadius',
        'minimum local difference in Number',
        'minimum local difference in AtomicWeight',
        'minimum local difference in Row',
        'minimum local difference in Column',
        'minimum local difference in Electronegativity',
        'minimum local difference in CovalentRadius' 
        ]



# -*- coding: utf-8 -*-

"""The utilities for ``cp_app``."""

import numpy as np
import yaml

Kb=1.3806504e-23  # Boltzmann constant in [J/K]
Ph=1.98644586e-23 # Planck constant in [J.cm]
Avo=6.02214076e23 # 1/mol
J2cal=0.2390
th2cm=33.35641

def read_vibspectrum(filename):
    frequencies=[]
    with open(filename) as fi:
        for line in fi.readlines()[3:-1]:
            frequencies.append(float(line.strip().split()[2]))
    return np.array(frequencies)

def read_frequencies_from_mesh(filename):
    mesh=yaml.load(open(filename))
    w=[fr["frequency"] for fr in mesh["phonon"][0]["band"]]
    w=np.array(w)*th2cm
    return w

def cv_from_pdos(temp, pdos):
    pdos=pdos[np.where(pdos[:,0]>0)]
    x = Ph * pdos[:,0] / Kb / temp
    expVal = np.exp(x)
    cv_contributions= np.sum(pdos[:,1:],axis=1)* Avo*Kb * x ** 2 * expVal / (expVal - 1.0) ** 2
    return np.sum(cv_contributions)

def cv_from_dos(temp, totaldos):
    dos=totaldos[np.where(totaldos[:,0]>0)]
    x = Ph * dos[:,0] / Kb / temp
    expVal = np.exp(x)
    cv_contributions= dos[:,1]* Avo*Kb * x ** 2 * expVal / (expVal - 1.0) ** 2
    return np.sum(cv_contributions)

def cv_from_frequencies(temp, freqs):
    freqs=freqs[freqs>0]
    x = Ph * freqs / Kb / temp
    expVal = np.exp(x)
    cv_contributions=Avo*Kb * x ** 2 * expVal / (expVal - 1.0) ** 2
    return np.sum(cv_contributions)

def read_totaldos(filename):
    data=np.loadtxt(filename,skiprows=1)
    data[:,0]*=th2cm
    return data

def read_pdos(filename):
    data=np.loadtxt(filename,skiprows=1)
    data[:,0]*=th2cm
    return data


def add_type_label(mydict,atomtype,name,label):
    if atomtype in mydict:
        mydict[atomtype][name]=label
    else:
        mydict[atomtype]={name:label}
    return mydict

def read_atoms_from_mesh(filename):
    mesh=yaml.load(open(filename))
    w=[fr["frequency"] for fr in mesh["phonon"][0]["band"]]
    w=np.array(w)*th2cm
    return w

def cv_from_pdos_site(temp, pdos,site):
    pdos=pdos[np.where(pdos[:,0]>0)]
    x = Ph * pdos[:,0] / Kb / temp
    expVal = np.exp(x)
    cv_contributions= pdos[:,site+1]* Avo*Kb * x ** 2 * expVal / (expVal - 1.0) ** 2
    return np.sum(cv_contributions)

def select_structures(nsamples,df):
    selected=set()
    for structure_type in df["structure_type"].unique():
        selected.add(df.loc[df["structure_type"]==structure_type].index.values[0])
        selected.add(df.loc[df["structure_type"]==structure_type].index.values[1])
        if len(selected)>nsamples-1:
            break
    for atom_type in df["atom_types"].unique():
        selected.add(df.loc[df["atom_types"]==atom_type].index.values[0])
        if len(selected)>nsamples-1:
            break
        
    while len(selected)< nsamples:
        selected.add(df.sample(1).index.values[0])
        
    return selected




# -*- coding: utf-8 -*-

"""The methods to featurize a dataset of porous materials for heat capacity."""

import string
from pathlib import Path
import numpy as np
import pandas as pd
from typing import List
from pymatgen.io.cif import CifParser
from matminer.featurizers.site import GaussianSymmFunc, SiteElementalProperty,AGNIFingerprints 
from matminer.featurizers.base import BaseFeaturizer
from matminer.utils.data import MagpieData
from pymatgen.analysis.local_env import VoronoiNN


def featurize_dataset(cifs: list, verbos=False, saveto: str="features.csv")-> pd.DataFrame:
    """Featurize crystal structures using elemetal, geometric, and chemical descriptors for local environments.

    :params cifs: list of paths to crystal structure in cif format
    :params verbos: printing the steps
    :params saveto: filename to save the generated features
    """

    features={}
    for cif in cifs:
        structure = CifParser(cif).get_structures()[0]
        structure_name = Path(cif).name
        features[structure_name]={}
        features[structure_name]["structure"]=structure
        features[structure_name]["structure_name"]=Path(cif).name
        features[structure_name]["structure_path"]=str(Path(cif).parent)


    data=pd.DataFrame.from_dict(features).T
    features_dict={}
    ## 1. initialize the dictionary for each site
    for index,row in data.iterrows():
        structure=row["structure"]
        for atomidx in range(structure.num_sites):
            site_name="%s_%i"%(index,atomidx)
            features_dict[site_name]={}
            features_dict[site_name]={"structure_name": row["structure_name"]}
            features_dict[site_name].update({"structure_path": row["structure_path"]})

    ## 2. Site Elemental Property
    print("site elemental properties")
    property_list=("Number","AtomicWeight","Row","Column","Electronegativity","CovalentRadius")
    SEP = SiteElementalProperty(properties=property_list)
    colnames=SEP._generate_column_labels(multiindex=False,return_errors=False)
    for index,row in data.iterrows():
        structure=row["structure"]
        if verbos:
            print(index)
        for atomidx in range(structure.num_sites):
            feat=SEP.featurize(structure,idx=atomidx)
            site_name="%s_%i"%(index,atomidx)
            features_dict[site_name].update(dict(zip(colnames, feat)))

    ## 3. AGNI
    print("AGNI")
    property_list=("Number","AtomicWeight","Row","Column","Electronegativity","CovalentRadius")
    AGNI = AGNIFingerprints(cutoff=5,directions=[None])
    colnames=AGNI._generate_column_labels(multiindex=False,return_errors=False)
    for index,row in data.iterrows():
        structure=row["structure"]
        if verbos:
            print(index)
        for atomidx in range(structure.num_sites):
            feat=AGNI.featurize(structure,idx=atomidx)
            site_name="%s_%i"%(index,atomidx)
            features_dict[site_name].update(dict(zip(colnames, feat)))


    ## 4. Gaussian Symmetry Functions 
    print("GSF")
    GSF = GaussianSymmFunc(cutoff=5)
    colnames=GSF._generate_column_labels(multiindex=False,return_errors=False)
    for index,row in data.iterrows():
        structure=row["structure"]
        if verbos:
            print(index)
        for atomidx in range(structure.num_sites):
            feat=GSF.featurize(structure,idx=atomidx)
            site_name="%s_%i"%(index,atomidx)
            features_dict[site_name].update(dict(zip(colnames, feat)))

    ## 5. site difference stats 
    print("LPD")
    LPD = LocalPropertyStatsNew(properties=property_list)
    colnames=LPD._generate_column_labels(multiindex=False,return_errors=False)
    for index,row in data.iterrows():
        structure=row["structure"]
        if verbos:
            print(index)
        for atomidx in range(structure.num_sites):
            feat=LPD.featurize(structure,idx=atomidx)
            site_name="%s_%i"%(index,atomidx)
            features_dict[site_name].update(dict(zip(colnames, feat)))


    df_features=pd.DataFrame.from_dict(features_dict).T
    if saveto:
        df_features.to_csv(saveto)

    return df_features


def featurize_structure(cif: string, verbos=False, saveto: str="features.csv")-> pd.DataFrame:
    """Featurize a crystal structure using elemetal, geometric, and chemical descriptors for local environments.

    :params cifs: list of paths to crystal structure in cif format
    :params verbos: printing the steps
    :params saveto: filename to save the generated features
    """
    structure = CifParser(cif).get_structures()[0]
    structure_name = Path(cif).name
    features = {structure_name:{}}
    features[structure_name]["structure"]=structure
    features[structure_name]["structure_name"]=Path(cif).name
    features[structure_name]["structure_path"]=str(Path(cif).parent)
    
    data=pd.DataFrame.from_dict(features).T
    features_dict={}
    ## 1. initialize the dictionary for each site
    for index,row in data.iterrows():
        structure=row["structure"]
        for atomidx in range(structure.num_sites):
            site_name="%s_%i"%(index,atomidx)
            features_dict[site_name]={"structure_name": row["structure_name"]}
            features_dict[site_name].update({"structure_path": row["structure_path"]})

    ## 2. Site Elemental Property
    print("site elemental properties")
    property_list=("Number","AtomicWeight","Row","Column","Electronegativity","CovalentRadius")
    SEP = SiteElementalProperty(properties=property_list)
    colnames=SEP._generate_column_labels(multiindex=False,return_errors=False)
    for index,row in data.iterrows():
        structure=row["structure"]
        if verbos:
            print(index)
        for atomidx in range(structure.num_sites):
            feat=SEP.featurize(structure,idx=atomidx)
            site_name="%s_%i"%(index,atomidx)
            features_dict[site_name].update(dict(zip(colnames, feat)))

    ## 3. AGNI
    print("AGNI")
    property_list=("Number","AtomicWeight","Row","Column","Electronegativity","CovalentRadius")
    AGNI = AGNIFingerprints(cutoff=5,directions=[None])
    colnames=AGNI._generate_column_labels(multiindex=False,return_errors=False)
    for index,row in data.iterrows():
        structure=row["structure"]
        if verbos:
            print(index)
        for atomidx in range(structure.num_sites):
            feat=AGNI.featurize(structure,idx=atomidx)
            site_name="%s_%i"%(index,atomidx)
            features_dict[site_name].update(dict(zip(colnames, feat)))


    ## 4. Gaussian Symmetry Functions 
    print("GSF")
    GSF = GaussianSymmFunc(cutoff=5)
    colnames=GSF._generate_column_labels(multiindex=False,return_errors=False)
    for index,row in data.iterrows():
        structure=row["structure"]
        if verbos:
            print(index)
        for atomidx in range(structure.num_sites):
            feat=GSF.featurize(structure,idx=atomidx)
            site_name="%s_%i"%(index,atomidx)
            features_dict[site_name].update(dict(zip(colnames, feat)))

    ## 5. site difference stats 
    print("LPD")
    LPD = LocalPropertyStatsNew(properties=property_list)
    colnames=LPD._generate_column_labels(multiindex=False,return_errors=False)
    for index,row in data.iterrows():
        structure=row["structure"]
        if verbos:
            print(index)
        for atomidx in range(structure.num_sites):
            feat=LPD.featurize(structure,idx=atomidx)
            site_name="%s_%i"%(index,atomidx)
            features_dict[site_name].update(dict(zip(colnames, feat)))


    df_features=pd.DataFrame.from_dict(features_dict).T
    if saveto:
        df_features.to_csv(saveto)

    return df_features


class LocalPropertyStatsNew(BaseFeaturizer):
    """
    Differences, minima and maxima in elemental properties between site and its neighboring sites.
    Uses the Voronoi tessellation of the structure to determine the
    neighbors of the site, and assigns each neighbor (:math:`n`) a
    weight (:math:`A_n`) that corresponds to the area of the facet
    on the tessellation corresponding to that neighbor.
    The local property difference is then computed by
    :math:`\\frac{\sum_n {A_n |p_n - p_0|}}{\sum_n {A_n}}`
    where :math:`p_n` is the property (e.g., atomic number) of a neighbor
    and :math:`p_0` is the property of a site. If signed parameter is assigned
    True, signed difference of the properties is returned instead of absolute
    difference.
    Taken from oximachine featurizer: https://github.com/kjappelbaum/oximachine_featurizer
    Features:
        - "local property stat in [property]"
    References:
         `Ward et al. _PRB_ 2017 <http://link.aps.org/doi/10.1103/PhysRevB.96.024104>`_
    """

    def __init__(
        self,
        data_source=MagpieData(),
        weight: str = "area",
        properties: List[str] = ("Electronegativity",),
        cutoff: List[str] = 5,
    ):
        """Initialize the featurizer
        Args:
            data_source (AbstractData) - Class from which to retrieve
                elemental properties
            weight (str) - What aspect of each voronoi facet to use to
                weigh each neighbor (see VoronoiNN)
            properties (List[str]) - List of properties to use (default=['Electronegativity'])
            cutoff (float)
        """
        self.data_source = data_source
        self.properties = properties
        self.weight = weight
        self.cutoff = cutoff

    @staticmethod
    def from_preset(preset: str, cutoff: float = 13):
        """
        Create a new LocalPropertyStats class according to a preset
        Args:
            preset (str) - Name of preset
            cutoff (float) - Cutoff for the nearest neighbor search
        """

        if preset == "interpretable":
            return LocalPropertyStatsNew(
                data_source=MagpieData(),
                properties=[
                    "MendeleevNumber",
                    "Column",
                    "Row",
                    "Electronegativity",
                    "NsValence",
                    "NpValence",
                    "NdValence",
                    "NfValence",
                    "NValence",
                    "NsUnfilled",
                    "NpUnfilled",
                    "NdUnfilled",
                    "NfUnfilled",
                    "NUnfilled",
                    "GSbandgap",
                ],
                cutoff=cutoff,
            )
        else:
            raise ValueError("Unrecognized preset: " + preset)

    def featurize(self, strc, idx: int):
        # Get the targeted site
        my_site = strc[idx]

        # Get the tessellation of a site
        nn = VoronoiNN(
            weight=self.weight,
            tol=0.0,
            cutoff=self.cutoff,
            compute_adj_neighbors=False,
        ).get_nn_info(strc, idx)

        # Get the element and weight of each site
        elems = [n["site"].specie for n in nn]
        weights = [n["weight"] for n in nn]

        # Compute the difference for each property
        output = np.zeros((len(self.properties),))
        output_signed = np.zeros((len(self.properties),))
        output_max = np.zeros((len(self.properties),))
        output_min = np.zeros((len(self.properties),))

        total_weight = np.sum(weights)
        for i, p in enumerate(self.properties):
            my_prop = self.data_source.get_elemental_property(my_site.specie, p)
            n_props = self.data_source.get_elemental_properties(elems, p)
            output[i] = (
                np.dot(weights, np.abs(np.subtract(n_props, my_prop))) / total_weight
            )
            output_signed[i] = (
                np.dot(weights, np.subtract(n_props, my_prop)) / total_weight
            )
            output_max[i] = np.max(np.subtract(n_props, my_prop))
            output_min[i] = np.min(np.subtract(n_props, my_prop))
        return np.hstack([output, output_signed, output_max, output_min])

    def feature_labels(self):

        return (
            ["local difference in " + p for p in self.properties]
            + ["local signed difference in " + p for p in self.properties]
            + ["maximum local difference in " + p for p in self.properties]
            + ["minimum local difference in " + p for p in self.properties]
        )

    def citations(self):
        return [
            "@article{Ward2017,"
            "author = {Ward, Logan and Liu, Ruoqian "
            "and Krishna, Amar and Hegde, Vinay I. "
            "and Agrawal, Ankit and Choudhary, Alok "
            "and Wolverton, Chris},"
            "doi = {10.1103/PhysRevB.96.024104},"
            "journal = {Physical Review B},"
            "pages = {024104},"
            "title = {{Including crystal structure attributes "
            "in machine learning models of formation energies "
            "via Voronoi tessellations}},"
            "url = {http://link.aps.org/doi/10.1103/PhysRevB.96.014107},"
            "volume = {96},year = {2017}}",
            "@article{jong_chen_notestine_persson_ceder_jain_asta_gamst_2016,"
            "title={A Statistical Learning Framework for Materials Science: "
            "Application to Elastic Moduli of k-nary Inorganic Polycrystalline Compounds}, "
            "volume={6}, DOI={10.1038/srep34256}, number={1}, journal={Scientific Reports}, "
            "author={Jong, Maarten De and Chen, Wei and Notestine, Randy and Persson, "
            "Kristin and Ceder, Gerbrand and Jain, Anubhav and Asta, Mark and Gamst, Anthony}, "
            "year={2016}, month={Mar}}",
        ]

    def implementors(self):
        return ["Logan Ward", "Aik Rui Tan"]


def add_type_feature(mydict,atomtype,name,feat):
    if atomtype in mydict:
        mydict[atomtype][name]=feat
    else:
        mydict[atomtype]={name:feat}
    return mydict



# -*- coding: utf-8 -*-

"""The methods to compute labels for machine learning the heat capacity."""

from ase.io import read
from .utils import  cv_from_pdos_site 
import pandas as pd
import numpy as np
import phonopy
from phonopy.units import CP2KToTHz

def compute_total_dos_structure(phonopy_params: str, unitfactor=CP2KToTHz, dx:float=0.5, fmax:float=100.0, freq_pitch:float=0.5, saveto: str=None):
    """Compute projected dos from phonopy parameter file

    :param phonopy_params: list of phonopy parameter files (output of DFT)
    :param cif: list of crystal structure in cif format
    :param temperatures: the target temperature 
    :param factor: the unit conversion factor
    :param dx: spacing to compute dos
    :param fmax: max frequency in dos calculations
    :param freq_pitch: pitch frequency in dos calculations
    """
    phonon = None
    phonon = phonopy.load(phonopy_params, factor=unitfactor)
    phonon.run_mesh([1,1,1], with_eigenvectors=True)
    # total dos
    phonon.run_total_dos(sigma=dx,freq_min=0,freq_max=fmax, freq_pitch=freq_pitch)
    dos_dict = phonon.get_total_dos_dict()
    totaldos = np.vstack([dos_dict["frequency_points"],dos_dict["total_dos"]]).T
    if saveto:
        np.savetxt(saveto,totaldos,header="sigma = {}".format(dx))

    return totaldos


def compute_projected_dos_structure(phonopy_params: str, unitfactor=CP2KToTHz, dx=0.5, fmax=100, freq_pitch=0.5, saveto: str=None):
    """Compute projected dos from phonopy parameter file

    :param phonopy_params: list of phonopy parameter files (output of DFT)
    :cif: list of crystal structure in cif format
    :temperatures: the target temperature 
    :factor: the unit conversion factor
    :dx: spacing to compute dos
    :fmax: max frequency in dos calculations
    :freq_pitch: pitch frequency in dos calculations
    :saveto: save the projected dos to a file
    """
    phonon = None
    phonon = phonopy.load(phonopy_params, factor=unitfactor)
    phonon.run_mesh([1,1,1], with_eigenvectors=True)
    mesh = phonon.get_mesh_dict()
    # pdos
    phonon.run_projected_dos(sigma=dx,freq_min=0,freq_max=fmax, freq_pitch=freq_pitch)
    pdos_dict = phonon.get_projected_dos_dict()
    pdos = np.vstack([pdos_dict["frequency_points"],pdos_dict["projected_dos"]]).T
    if saveto:
        np.savetxt(saveto, pdos, header="sigma = {}".format(dx))

    return pdos 

def compute_atomic_cv_dataset(phonopy_params: list[str], cifs: list[str], temperatures: list, verbos=False, saveto: str="labels.csv") -> pd.DataFrame:
    """Compute atomic contribution to total cv from phonopy parameter file

    :param phonopy_params: list of phonopy parameter files (output of DFT)
    :cifs: list of crystal structure in cif format
    :temperatures: the target temperature 
    """
    labels={}
    for phonopy_param,cif in zip(phonopy_params,cifs):
        if verbos:
            print(cif)
        pdos=compute_projected_dos_structure(phonopy_params=phonopy_param, saveto="%s_projected_dos.dat".format(cif.replace(".cif","")))
        pdos[:,0]*=th2cm
        atoms=read(cif)
        if not atoms.get_global_number_of_atoms()==pdos.shape[1]-1:
            print(atoms.get_global_number_of_atoms(),pdos.shape[0])
            print("Warning! number of atoms do not match in pdos and structure for %s"%cif)
            continue
        for atomidx in range(atoms.get_global_number_of_atoms()):
            site_name="%s_%i"%(cif.replace(".cif",""),atomidx)
            for i,temperature in enumerate(temperatures):
                cv_site=cv_from_pdos_site(temperature,pdos,atomidx)
                if i==0:
                    labels[site_name]={"pCv_%05.2f"%temperature:cv_site}
                else:
                    labels[site_name]["pCv_%05.2f"%temperature]=cv_site

    df=pd.DataFrame.from_dict(labels).T
    df.to_csv(saveto)

    return df



try:
    from ccdc import io
    csd_reader = io.EntryReader('CSD')
except:
    print("You need to install CSD software with the license if you want to download all structures")

import os, json, requests, zipfile
from ase.io import read
from pymatgen.io.ase import AseAtomsAdaptor

from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

package_directory = os.path.abspath(__file__).replace("structure.py","")

files_to_download = {
                    'data/detail_of_CR.json': 'https://raw.githubusercontent.com/mtap-research/CoRE-MOF-Tools/main/7-data4API/detail_of_CR.json',
                    'data/detail_of_NCR.json': 'https://raw.githubusercontent.com/mtap-research/CoRE-MOF-Tools/main/7-data4API/detail_of_NCR.json',
                    'data/CSD/CR_CSD_REFCODE.json': 'https://raw.githubusercontent.com/mtap-research/CoRE-MOF-Tools/main/7-data4API/CSD/CR_CSD_REFCODE.json',
                    'data/CSD/NCR_CSD_REFCODE.json': 'https://raw.githubusercontent.com/mtap-research/CoRE-MOF-Tools/main/7-data4API/CSD/NCR_CSD_REFCODE.json',
                    'data/SI/CR.zip': 'https://raw.githubusercontent.com/mtap-research/CoRE-MOF-Tools/main/7-data4API/SI/CR.zip',
                    'data/SI/NCR.zip': 'https://raw.githubusercontent.com/mtap-research/CoRE-MOF-Tools/main/7-data4API/SI/NCR.zip'
                    }

for file_name, url in files_to_download.items():
    file_path = os.path.join(package_directory, file_name)
    directory = os.path.dirname(file_path) 

    if not os.path.exists(directory):
        os.makedirs(directory)
        
    if not os.path.exists(file_path):
        response = requests.get(url)
        if response.status_code == 200:
            with open(file_path, 'wb') as f:
                f.write(response.content)
            print(f"Downloaded {file_name} to {file_path}")
        else:
            print(f"Failed to download {file_name} from {url}")
    else:
        pass


def make_primitive_p1(filename):

    atoms = read(filename)
    structure_= AseAtomsAdaptor.get_structure(atoms)
    
    sga = SpacegroupAnalyzer(structure_)
    structure_prep = sga.get_primitive_standard_structure(international_monoclinic=True, keep_site_properties=False)
    structure_prep.to(filename=filename)
    

class download_from_SI():
    def __init__(self, output_folder="./CoREMOF2024DB"):
        
        self.SI_path = package_directory+'/data/SI/'
        self.output = output_folder
        self.run()

    def run(self):

        CR_files = self.list_zip(self.SI_path+"CR.zip")
        NCR_files = self.list_zip(self.SI_path+"NCR.zip")
     
        os.makedirs(self.output+"/CR/", exist_ok=True)
        os.makedirs(self.output+"/NCR/", exist_ok=True)

        for file in CR_files[:]:
            self.get_from_SI(self.SI_path+"CR.zip", file, self.output)
        for file in NCR_files[:]:
            self.get_from_SI(self.SI_path+"NCR.zip", file, self.output)

    def list_zip(self, zip_path):
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            file_list = zip_ref.namelist()
            return file_list
    
    def get_from_SI(self, zip_path, entry, output_folder):
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            file_list = zip_ref.namelist()
            if entry in file_list:
                zip_ref.extract(entry, output_folder)
            
def download_from_CSD(refcode, output_folder="./CoREMOF2024DB"):
    cryst = csd_reader.crystal(refcode)
    data = cryst.to_string('cif')
    f = open(os.path.join(output_folder, refcode+'.cif'),'w')
    f.write(data)
    f.close()

def get_list_CSD():

    CSD_path = package_directory+'/data/CSD/'
    CR_json = CSD_path + "/CR_CSD_REFCODE.json"
    NCR_json  = CSD_path + "/NCR_CSD_REFCODE.json"
    
    with open (CR_json, "r") as CR_f:
        CSD_CR = json.load(CR_f)
        
    with open (NCR_json, "r") as NCR_f:
        CSD_NCR = json.load(NCR_f)
    return CSD_CR, CSD_NCR


def information(dataset, entry):

    CR_data_path = package_directory+'/data/detail_of_CR.json'
    NCR_data_path = package_directory+'/data/detail_of_NCR.json'

    with open (CR_data_path, "r") as CR_f:
        CR_data = json.load(CR_f)
        
    with open (NCR_data_path, "r") as NCR_f:
        NR_data = json.load(NCR_f)
    
    if dataset == "CR-ASR":
        print("unit:\n", CR_data["unit"])
        query_data = CR_data["ASR"]
    elif dataset == "CR-FSR":
        print("unit:\n", CR_data["unit"])
        query_data = CR_data["FSR"]
    elif dataset == "CR-Ion":
        print("unit:\n", CR_data["unit"])
        query_data = CR_data["Ion"]
    elif dataset == "NCR":
        query_data = NR_data

    return query_data[entry]


import os
import subprocess

def ChanDim(structure, probe_radius = 0, high_accuracy = True):
    results_chan = {}
    results_chan["unit"]="nan"
    
    if high_accuracy:
        cmd = f'network -ha -chan {probe_radius} tmp_chan.txt {structure}'
    else:
        cmd = f'network -chan {probe_radius} tmp_chan.txt {structure}'
    _ = subprocess.run(
                        cmd,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        check=True,
                    )

    with open('tmp_chan.txt') as f:
        for i, row in enumerate(f):
            if i == 0:
                dim = int(row.split('dimensionality')[1].split()[0])

    results_chan["Dimention"] = dim

    os.remove("tmp_chan.txt")

    return results_chan

def FrameworkDim(structure, high_accuracy = True):
    results_strinfo = {}
    results_strinfo["unit"]="nan"
    
    if high_accuracy:
        cmd = f'network -ha -strinfo tmp_strinfo.txt {structure}'
    else:
        cmd = f'network -strinfo tmp_strinfo.txt {structure}'
    _ = subprocess.run(
                        cmd,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        check=True,
                    )

    with open('tmp_strinfo.txt') as f:
        line = f.readline().split()
        dim = int(line[-1])
        one_dim = int(line[7])
        two_dim = int(line[8])
        three_dim = int(line[9])
    results_strinfo["Dimention"] = dim
    results_strinfo["N_1D"] = one_dim
    results_strinfo["N_2D"] = two_dim
    results_strinfo["N_3D"] = three_dim

    os.remove("tmp_strinfo.txt")

    return results_strinfo

def PoreDiameter(structure, high_accuracy = True):

    results_pd = {}
    results_pd["unit"]="angstrom, Å"
    
    if high_accuracy:
        cmd = f'network -ha -res tmp_pd.txt {structure}'
    else:
        cmd = f'network -res tmp_pd.txt {structure}'
    _ = subprocess.run(
                        cmd,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        check=True,
                    )
    with open('tmp_pd.txt') as f:
        line = f.readline().split()
        results_pd["LCD"], results_pd["PLD"], results_pd["LFPD"] = map(float, line[1:4])
    os.remove('tmp_pd.txt')

    return results_pd

def SurfaceArea(structure, chan_radius = 1.655, probe_radius = 1.655, num_samples = 5000, high_accuracy = True):
    results_sa = {}
    results_sa["unit"]="Å^2, m^2/cm^3, m^2/g"
    
    if high_accuracy:
        cmd = f'network -ha -sa {chan_radius} {probe_radius} {num_samples} tmp_sa.txt {structure}'
    else:
        cmd = f'network -sa {chan_radius} {probe_radius} {num_samples} tmp_sa.txt {structure}'
    _ = subprocess.run(
                        cmd,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        check=True,
                    )
    with open('tmp_sa.txt') as f:
        for i, row in enumerate(f):
            if i == 0:
                ASA = float(row.split('ASA_A^2:')[1].split()[0])
                VSA = float(row.split('ASA_m^2/cm^3:')[1].split()[0])
                GSA = float(row.split('ASA_m^2/g:')[1].split()[0])
                NASA = float(row.split('NASA_A^2:')[1].split()[0])
                NVSA = float(row.split('NASA_m^2/cm^3:')[1].split()[0])
                NGSA = float(row.split('NASA_m^2/g:')[1].split()[0])

    results_sa["ASA"] = [ASA, VSA, GSA]
    results_sa["NASA"] = [NASA, NVSA, NGSA]

    os.remove("tmp_sa.txt")

    return results_sa

def PoreVolume(structure, chan_radius = 0, probe_radius = 0, num_samples = 5000, high_accuracy = True):
    results_pv = {}
    results_pv["unit"]="PV: Å^3, cm^3/g; VF: nan"
    
    if high_accuracy:
        cmd = f'network -ha -volpo {chan_radius} {probe_radius} {num_samples} tmp_pv.txt {structure}'
    else:
        cmd = f'network -volpo {chan_radius} {probe_radius} {num_samples} tmp_pv.txt {structure}'
    _ = subprocess.run(
                        cmd,
                        shell=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        check=True,
                    )
    with open('tmp_pv.txt') as f:
        for i, row in enumerate(f):
            if i == 0:
                POAV = float(row.split('POAV_A^3:')[1].split()[0])
                PONAV = float(row.split('PONAV_A^3:')[1].split()[0])
                GPOAV = float(row.split('POAV_cm^3/g:')[1].split()[0])
                GPONAV = float(row.split('PONAV_cm^3/g:')[1].split()[0])
                POAV_volume_fraction = float(row.split('POAV_Volume_fraction:')[1].split()[0])
                PONAV_volume_fraction = float(row.split('PONAV_Volume_fraction:')[1].split()[0])
    results_pv["PV"] = [POAV, GPOAV]
    results_pv["NPV"] = [PONAV, GPONAV]
    results_pv["VF"] = POAV_volume_fraction
    results_pv["NVF"] = PONAV_volume_fraction

    os.remove("tmp_pv.txt")

    return results_pv



import os, re, csv, glob, functools, warnings, itertools, collections
from ase.io import read, write
from pymatgen.core import Structure
from pymatgen.io.cif import CifWriter

from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from CoREMOF.utils.atoms_definitions import METAL, COVALENTRADII
from CoREMOF.utils.ions_list import ALLIONS
import numpy as np
import pandas as pd

from ase.neighborlist import NeighborList
from scipy.sparse.csgraph import connected_components
warnings.filterwarnings('ignore')

import json

from pymatgen.io.ase import AseAtomsAdaptor
from mofchecker import MOFChecker
from CoREMOF.utils.atoms_definitions import ATR, Coef_A, Coef_C #, BO_list, metals4check


class preprocess():

    def __init__(self, structure, output_folder="result_curation"):
        self.structure = structure
        self.output = output_folder + os.sep
        os.makedirs(self.output, exist_ok=True)
        self.result_check=self.process()
                
    def process(self):
        result_check = self.split_pri_p1(self.structure, self.output)
        with open(self.output + os.path.basename(self.structure).replace(".cif","") + "_precheck.json", "w") as f:
            json.dump(result_check,f,indent=2)


    def split_pri_p1(self, structure, output_folder):
        result_check = {}
        
        structures = read(structure, index=':')
        n_struc = len(structures)
        result_check["N_structures"] = n_struc

        if n_struc > 1:
            if not os.path.exists(output_folder):
                os.makedirs(output_folder, exist_ok=True)
            print(structure, "with more than one crystal structures")
            for i, atoms in enumerate(structures):
                
                structure_= AseAtomsAdaptor.get_structure(atoms)
                sga = SpacegroupAnalyzer(structure_)
                structure_prep = sga.get_primitive_standard_structure(international_monoclinic=True, keep_site_properties=False)
                struct_name = os.path.basename(structure).replace(".cif","")+"_"+str(i+1)
                structure_prep.to(filename=os.path.join(output_folder, f"{struct_name}.cif"))

                has_metal = any(METAL.get(atom.symbol) for atom in atoms)
                has_carbon = any(atom.symbol == 'C' for atom in atoms)
                
                if has_metal:
                    if has_carbon:
                        result_check[struct_name] = "has metal and carbon"
                    else:
                        result_check[struct_name] = "missing carbon"
                else:
                    if has_carbon:
                        result_check[struct_name] = "missing metal"
                    else:
                        result_check[struct_name] = "missing metal and carbon"
        else:
            atoms = structures[0]
            struct_name = os.path.basename(structure).replace(".cif","")

            has_metal = any(METAL.get(atom.symbol) for atom in atoms)
            has_carbon = any(atom.symbol == 'C' for atom in atoms)
            
            if has_metal:
                if has_carbon:
                    result_check[struct_name] = "has metal and carbon"
                else:
                    result_check[struct_name] = "missing carbon"
            else:
                if has_carbon:
                    result_check[struct_name] = "missing metal"
                else:
                    result_check[struct_name] = "missing metal and carbon"

            structure_= AseAtomsAdaptor.get_structure(atoms)
            sga = SpacegroupAnalyzer(structure_)
            structure_prep = sga.get_primitive_standard_structure(international_monoclinic=True, keep_site_properties=False)
            temp_path = os.path.join(output_folder, f"{struct_name}.cif")
            structure_prep.to(filename=temp_path)
            
        return result_check
    

class clean():
    def __init__(self, structure, initial_skin = 0.25, output_folder="result_curation", saveto: str="clean_result.csv")-> pd.DataFrame:
        
        self.cambridge_radii = COVALENTRADII
        self.metal_list = [element for element, is_metal in METAL.items() if is_metal]
        self.ions_list = set(ALLIONS)
        
        self.structure = structure
        self.initial_skin = initial_skin
        self.output = output_folder
        os.makedirs(self.output, exist_ok=True)

        if saveto:
            self.csv_path = os.path.join(self.output, saveto)
        self.process()

    def process(self):

        try:
            with open(self.csv_path, mode='w', newline='') as csv_file:
                csv_writer = csv.writer(csv_file)
                csv_writer.writerow(["Name", "Skin_FSR", "Skin_ASR", "Removed_FSR", "Removed_ASR"])

                fsr_skin, fsr_results = self.run_fsr(self.structure, self.output, self.initial_skin, self.metal_list, self.ions_list)
                print(f"FSR results for {self.structure}: {fsr_results}")
                
                asr_skin, asr_results = self.run_asr(self.structure, self.output, self.initial_skin, self.metal_list, self.ions_list)
                print(f"ASR results for {self.structure}: {asr_results}")

                csv_writer.writerow([
                                        os.path.basename(self.structure),
                                        str(fsr_skin) if fsr_skin else "none",
                                        str(asr_skin) if asr_skin else "none",
                                        "; ".join(fsr_results if fsr_results else "none"),
                                        "; ".join(asr_results if asr_results else "none")
                                    ])
        except:
            fsr_skin, fsr_results = self.run_fsr(self.structure, self.output, self.initial_skin, self.metal_list, self.ions_list)
            print(f"FSR results for {self.structure}: {fsr_results}")
            
            asr_skin, asr_results = self.run_asr(self.structure, self.output, self.initial_skin, self.metal_list, self.ions_list)
            print(f"ASR results for {self.structure}: {asr_results}")

    def run_fsr(self, mof, save_folder, initial_skin, metal_list, ions_list):
        m = mof.replace(".cif", "")
        skin = initial_skin
        try:
            while True:
                skin, printed_formulas = self.free_clean(m, save_folder, ions_list, skin)
                has_metals = False
                for e_s in printed_formulas:
                    split_formula = re.findall(r'([A-Z][a-z]?)(\d*)', e_s)
                    elements = [match[0] for match in split_formula]
                    if any(e in metal_list for e in elements):
                        has_metals = True
                        skin += 0.05
                if not has_metals:
                    break
            return skin, printed_formulas
        except Exception as e:
            print(m, str(e))


    def run_asr(self, mof, save_folder, initial_skin, metal_list, ions_list):
        m = mof.replace(".cif", "")
        skin = initial_skin
        try:
            while True:
                skin, printed_formulas = self.all_clean(m, save_folder, ions_list, skin)
                has_metals = False
                for e_s in printed_formulas:
                    split_formula = re.findall(r'([A-Z][a-z]?)(\d*)', e_s)
                    elements = [match[0] for match in split_formula]
                    if any(e in metal_list for e in elements):
                        has_metals = True
                        skin += 0.05
                if not has_metals:
                    break
            return skin, printed_formulas
        except Exception as e:
            print(f"{m} Fail: {e}")

    def build_ASE_neighborlist(self, cif, skin):
        radii = [self.cambridge_radii[i] for i in cif.get_chemical_symbols()]
        ASE_neighborlist = NeighborList(radii, self_interaction=False, bothways=True, skin=skin)
        ASE_neighborlist.update(cif)
        return ASE_neighborlist

    def find_clusters(self, adjacency_matrix, atom_count):
        clusters = []
        cluster_count, clusterIDs = connected_components(adjacency_matrix, directed=True)
        for n in range(cluster_count):
            clusters.append([i for i in range(atom_count) if clusterIDs[i] == n])
        return clusters

    def find_metal_connected_atoms(self, structure, neighborlist):
        metal_connected_atoms = []
        metal_atoms = []
        for i, elem in enumerate(structure.get_chemical_symbols()):
            if elem in self.metal_list:
                neighbors, _ = neighborlist.get_neighbors(i)
                metal_connected_atoms.append(neighbors)
                metal_atoms.append(i)
        return metal_connected_atoms, metal_atoms, structure

    def CustomMatrix(self, neighborlist, atom_count):
        matrix = np.zeros((atom_count, atom_count), dtype=int)
        for i in range(atom_count):
            neighbors, _ = neighborlist.get_neighbors(i)
            for j in neighbors:
                matrix[i][j] = 1
        return matrix

    def mod_adjacency_matrix(self, adj_matrix, MetalConAtoms, MetalAtoms, atom_count, struct):
        clusters = self.find_clusters(adj_matrix, atom_count)
        for i, element_1 in enumerate(MetalAtoms):
            for j, element_2 in enumerate(MetalConAtoms[i]):
                if struct[element_2].symbol == "O":
                    tmp = len(self.find_clusters(adj_matrix, atom_count))
                    adj_matrix[element_2][element_1] = 0
                    adj_matrix[element_1][element_2] = 0
                    new_clusters = self.find_clusters(adj_matrix, atom_count)
                    if tmp == len(new_clusters):
                        adj_matrix[element_2][element_1] = 1
                        adj_matrix[element_1][element_2] = 1
                    for ligand in new_clusters:
                        if ligand not in clusters:
                            tmp3 = struct[ligand].get_chemical_symbols()
                            if "O" and "H" in tmp3 and len(tmp3) == 2:
                                adj_matrix[element_2][element_1] = 1
                                adj_matrix[element_1][element_2] = 1
        return adj_matrix

    def cmp(self, x, y):
        return (x > y) - (x < y)

    def cluster_to_formula(self, cluster, cif):
        symbols = [cif[i].symbol for i in cluster]
        count = collections.Counter(symbols)
        formula = ''.join([atom + (str(count[atom]) if count[atom] > 1 else '') for atom in sorted(count)])
        return formula

    def free_clean(self, input_file, save_folder, ions_list, skin):
        try:
            print(input_file+".cif")
            cif = read(input_file+".cif")
            refcode = input_file.split("/")[-1]
            atom_count = len(cif.get_chemical_symbols())
            ASE_neighborlist = self.build_ASE_neighborlist(cif,skin)
            a = self.CustomMatrix(ASE_neighborlist,atom_count)
            b = self.find_clusters(a,atom_count)
            b.sort(key=functools.cmp_to_key(lambda x,y: self.cmp(len(x), len(y))))
            b.reverse()
            cluster_length=[]
            solvated_cluster = []
            ions_cluster = []
            printed_formulas = []
            iii=False
            print("using",skin,"as skin")
            for index, _ in enumerate(b):
                cluster_formula = self.cluster_to_formula(b[index], cif) 
                if cluster_formula in ions_list:
                    print(cluster_formula, "is ion")
                    ions_cluster.append(b[index])
                    iii=True
                else:
                    tmp = len(b[index])
                    if len(cluster_length) > 0:
                        if tmp > max(cluster_length):
                            cluster_length = []
                            solvated_cluster = []
                            solvated_cluster.append(b[index])
                            cluster_length.append(tmp)
                        elif tmp > 0.5 * max(cluster_length):
                            solvated_cluster.append(b[index])
                            cluster_length.append(tmp)
                        else:
                            formula = self.cluster_to_formula(b[index], cif)
                            if formula not in printed_formulas:
                                printed_formulas.append(formula)
                    else:
                        solvated_cluster.append(b[index])
                        cluster_length.append(tmp)
            solvated_cluster = solvated_cluster + ions_cluster
            solvated_merged = list(itertools.chain.from_iterable(solvated_cluster))
            atom_count = len(cif[solvated_merged].get_chemical_symbols())

            if iii:
                refcode=refcode.replace("_FSR","")
                new_fn = refcode + '_ION_FSR.cif'
            else:
                refcode=refcode.replace("_FSR","")
                new_fn = refcode + '_FSR.cif'
            write(os.path.join(save_folder, new_fn), cif[solvated_merged])
        
            return skin, printed_formulas
        except Exception as e:
            print(f"{input_file} Fail: {e}")

    def all_clean(self, input_file, save_folder, ions_list, skin):
        try:
            fn = input_file
            cif = read(input_file+".cif")
            refcode = fn.split("/")[-1]
            atom_count = len(cif.get_chemical_symbols())
            ASE_neighborlist = self.build_ASE_neighborlist(cif,skin)
            a = self.CustomMatrix(ASE_neighborlist,atom_count)
            b = self.find_clusters(a,atom_count)
            b.sort(key=functools.cmp_to_key(lambda x,y: self.cmp(len(x), len(y))))
            b.reverse()
            cluster_length=[]
            solvated_cluster = []

            printed_formulas = []

            ions_cluster = []
            iii=False
            print("using",skin,"as skin")

            for index, _ in enumerate(b):
                
                cluster_formula = self.cluster_to_formula(b[index], cif) 
                if cluster_formula in ions_list:
                    print(cluster_formula, "is ion")
                    ions_cluster.append(b[index])
                    solvated_cluster.append(b[index])
                    iii=True

                else:
                    tmp = len(b[index])
                    if len(cluster_length) > 0:
                        if tmp > max(cluster_length):
                            cluster_length = []
                            solvated_cluster = []
                            solvated_cluster.append(b[index])
                            cluster_length.append(tmp)
                        if tmp > 0.5 * max(cluster_length):
                            solvated_cluster.append(b[index])
                            cluster_length.append(tmp)
                        else:
                            formula = self.cluster_to_formula(b[index], cif)
                            if formula not in printed_formulas:
                                printed_formulas.append(formula)
                    else:
                        solvated_cluster.append(b[index])
                        cluster_length.append(tmp)
                    
            solvated_merged = list(itertools.chain.from_iterable(solvated_cluster))
            
            atom_count = len(cif[solvated_merged].get_chemical_symbols())
            
            newASE_neighborlist = self.build_ASE_neighborlist(cif[solvated_merged],skin)
            MetalCon, MetalAtoms, struct = self.find_metal_connected_atoms(cif[solvated_merged], newASE_neighborlist)
            c = self.CustomMatrix(newASE_neighborlist,atom_count)
            d = self.mod_adjacency_matrix(c, MetalCon, MetalAtoms,atom_count,struct)
            solvated_clusters2 = self.find_clusters(d,atom_count)
            solvated_clusters2.sort(key=functools.cmp_to_key(lambda x,y: self.cmp(len(x), len(y))))
            solvated_clusters2.reverse()
            cluster_length=[]
            final_clusters = []

            ions_cluster2 = []
            for index, _ in enumerate(solvated_clusters2):

                cluster_formula2 = self.cluster_to_formula(solvated_clusters2[index], struct) 
                if cluster_formula2 in ions_list:
                    final_clusters.append(solvated_clusters2[index])
                    iii=True
                else:
                    tmp = len(solvated_clusters2[index])
                    if len(cluster_length) > 0:
                        if tmp > max(cluster_length):
                            cluster_length = []
                            final_clusters = []
                            final_clusters.append(solvated_clusters2[index])
                            cluster_length.append(tmp)
                        if tmp > 0.5 * max(cluster_length):
                            final_clusters.append(solvated_clusters2[index])
                            cluster_length.append(tmp)
                        else:
                            formula = self.cluster_to_formula(solvated_clusters2[index], struct)
                            if formula not in printed_formulas:
                                printed_formulas.append(formula)
                    else:
                        final_clusters.append(solvated_clusters2[index])
                        cluster_length.append(tmp)
            if iii:
                final_clusters = final_clusters+ions_cluster2
            else:
                final_clusters = final_clusters
            final_merged = list(itertools.chain.from_iterable(final_clusters))
            tmp = struct[final_merged].get_chemical_symbols()
            tmp.sort()
            if iii:
                new_fn = refcode + '_ION_ASR.cif'
            else:
                new_fn = refcode + "_ASR.cif"
            write(os.path.join(save_folder, new_fn), struct[final_merged])
            return skin, printed_formulas
    
        except Exception as e:
            print(f"{input_file} Fail: {e}")

class mof_check():

    def __init__(self, structure, output_folder="result_curation"):
        self.output = output_folder + os.sep
        self.structure = structure
        os.makedirs(self.output, exist_ok=True)
        self.check()

    def check(self):
        result_check = {}

        chen_manz_result = self.Chen_Manz(self.structure)
        mof_checker_result = self.mof_checker(self.structure)

        result_check["Chen_Manz"] = chen_manz_result
        result_check["mofchecker"] = mof_checker_result

        with open(self.output + os.path.basename(self.structure).replace(".cif","") + "_Chen_Manz_mofchecker.json", "w") as f:
            json.dump(result_check,f,indent=2)

    def Chen_Manz(self, structure):
        try:
            
            has_problem =[]

            atoms = read(structure)
            sym = atoms.get_chemical_symbols()
            
            for a in range(len(atoms)):
                H_connected = []
                nl = []
                for b in range(len(atoms)):
                    if a == b:
                        continue
                    d = atoms.get_distance(a, b, mic = True)
                    if sym[a] == 'H':
                        if d <= (0.3 + ATR[sym[a]] + ATR[sym[b]]):
                            H_connected.append(b)
                    if d < 0.5*(ATR[sym[a]] + ATR[sym[b]]):
                        has_problem.append("overlapping")
                    if d <= (ATR[sym[a]] + ATR[sym[b]]):
                        nl.append(b)
                if sym[a] == 'C':
                    bonded_ele = [sym[e] for e in nl if sym[e] not in Coef_A]
                    if len(bonded_ele) != 0:
                        pass
                    else:
                        BO = []
                        for bidx in range(len(nl)):
                            b = nl[bidx]
                            d = atoms.get_distance(a, b, mic = True)
                            BO_ab = 10**(Coef_A[sym[b]]*d + Coef_C[sym[b]])
                            if sym[b] == 'H':
                                if BO_ab > 1.25:
                                    BO_ab = 1.25
                            BO.append(BO_ab)
                        sum_BO = sum(BO)
                        if sum_BO < 3.3:
                            has_problem.append("under_carbon")
                        elif sum_BO >= 5.5:
                            has_problem.append("over_carbon")
                if len(nl) == 0:
                    has_problem.append("isolated")

            if len(has_problem) > 0:
                return list(set(has_problem))
            else:
                return ["no issues"]
        except Exception as e:
            return ["unknown"]
        

    def mof_checker(self, structure):
        
        try:
            atoms = read(structure)
            structure_ = AseAtomsAdaptor.get_structure(atoms)
            checker = MOFChecker(structure_)
            check_result = checker.get_mof_descriptors()

            has_problem = []
            problem_keys_true = [
                "has_atomic_overlaps", "has_overcoordinated_c", "has_overcoordinated_n",
                "has_overcoordinated_h", "has_suspicious_terminal_oxo",
                "has_undercoordinated_c", "has_undercoordinated_n",
                "has_undercoordinated_rare_earth", "has_undercoordinated_alkali_alkaline",
                "has_geometrically_exposed_metal", 
                "has_lone_molecule"
            ]
            problem_keys_false = ["has_metal", "has_carbon"]

            for key in problem_keys_true:
                if check_result.get(key, False):
                    has_problem.append(key)

            for key in problem_keys_false:
                if not check_result.get(key, True):
                    has_problem.append(key)

            if len(has_problem) > 0:
                return list(set(has_problem))
            else:
                return ["no issues"]
        except Exception as e:
            return ["unknown"]



import numpy as np
import pandas as pd
from .descriptors import cv_features
import joblib
import glob
import copy

FEATURES = cv_features

def predict_Cv_ensemble_structure(models: list, FEATURES: list, df_features: pd.DataFrame, temperature: float) -> list:
  
    if len(df_features["structure_name"].unique())>2:
        raise ValueError("More than one structure in the features file...")
        
    df_site_structure = copy.deepcopy(df_features)
    structure_name = df_site_structure["structure_name"].unique()[0]
    print(structure_name)
    for model_idx,model in enumerate(models):
        df_site_structure["pCv_{}_predicted_{}".format(temperature, model_idx)]=model.predict(df_site_structure[FEATURES])
    results=[]
    predicted_mol=[]
    predicted_gr=[]
    for model_idx in range(len(models)):
        sites=df_site_structure.loc[df_site_structure["structure_name"]==structure_name]
        predicted_mol.append(np.sum(sites["pCv_{}_predicted_{}".format(temperature,model_idx)])/len(sites))
        predicted_gr.append(np.sum(sites["pCv_{}_predicted_{}".format(temperature,model_idx)])/np.sum(sites["site AtomicWeight"]))
    results.append({
        "name": structure_name,
        "Cv_gravimetric_{}_mean".format(temperature): np.mean(predicted_gr),
        "Cv_gravimetric_{}_std".format(temperature): np.std(predicted_gr),
        "Cv_molar_{}_mean".format(temperature): np.mean(predicted_mol),
        "Cv_molar_{}_std".format(temperature): np.std(predicted_mol),
    })
    return results


def predict_Cv_ensemble_structure_multitemperatures(path_to_models: str, structure_name: str, features_file: str="features.csv", 
                FEATURES: list=cv_features, temperatures: list=[300.00], save_to: str="cv_predicted.csv") -> pd.DataFrame:

    df_features = pd.read_csv(features_file)
    df_features["structure_name"]=["_".join(n.split("_")[:-1]) for n in df_features["Unnamed: 0"]]
    df_features=df_features.loc[df_features["structure_name"]==structure_name]
    for i,temperature in enumerate(temperatures):
        models=[]
        print("loading models for:", temperature)
        modelnames = glob.glob("{}/{:.0f}/*".format(path_to_models, temperature))
        models = [joblib.load(n) for n in modelnames]
        print("{} models loaded, predicting...".format(len(models)))
        if i==0:
            res= pd.DataFrame(predict_Cv_ensemble_structure(models, FEATURES, df_features, temperature))
            all_results=res
        else:
            res= pd.DataFrame(predict_Cv_ensemble_structure(models, FEATURES, df_features, temperature))
            all_results=all_results.merge(res, how="inner",on="name")

    if save_to:
        all_results.to_csv(save_to)
    return all_results


from setuptools import setup, find_packages

setup(
        name='CoREMOF_tools',
        version='0.0.2',
        author='Guobin Zhao',
        author_email='sxmzhaogb@gmail.com',
        description='Python API for CoRE MOF 2024 DB',
        long_description=open('README.md').read(),
        long_description_content_type='text/markdown',
        url='https://github.com/mtap-research/CoRE-MOF-Tools',
        packages=find_packages(),
        # package_data={
        # 'CoREMOF': [
        #     'data/*.json',
        #     'data/CSD/*.json',
        #     'data/SI/*.zip',
        #     'models/cp_app/ensemble_models_smallML_120_100/300/*',
        #     'models/cp_app/ensemble_models_smallML_120_100/350/*',
        #     'models/cp_app/ensemble_models_smallML_120_100/400/*',
        #     'models/stability/*'
        # ],
        # },
        install_requires=[
            'pymatgen',
            'ase',
            'juliacall',
            'molSimplify',
            'PACMAN-charge',
            'cloudpickle',
            'matminer',
            'xgboost',
            'scikit-learn==1.3.2',
            'mofchecker'
        ],
        extras_require={
            'zeopp': ['zeopp-lsmo']
        },
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'Topic :: Software Development :: Build Tools',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.9',
        ],
        python_requires='>=3.9, <4',
        # entry_points={
        #     'console_scripts': [
        #         'coremof=CoREMOF:curate',
            # ],
        # },
    )



from PACMANCharge import pmcharge

import os, warnings, shutil, joblib, requests, cloudpickle
import keras
import keras.backend as K
import pandas as pd

warnings.filterwarnings('ignore')

import numpy as np
import pickle as pkl

from CoREMOF.calculation import Zeopp
from CoREMOF.calculation.mof_features import RACs, Volume

package_directory = os.path.abspath(__file__).replace("prediction.py","")

def get_files_from_github(repo, path):
    url = f"https://api.github.com/repos/{repo}/contents/{path}"
    headers = {'Accept': 'application/vnd.github.v3+json'}
    response = requests.get(url, headers=headers)
    response.raise_for_status() 
    return response.json()

def download_file(url, save_path):
    if not os.path.exists(save_path):
        response = requests.get(url)
        response.raise_for_status()
        os.makedirs(os.path.dirname(save_path), exist_ok=True) 
        with open(save_path, 'wb') as file:
            file.write(response.content)
        print(f"Downloaded {save_path}")
    else:
        pass

repo = 'mtap-research/CoRE-MOF-Tools'
github_paths = [
    '4-MLFeatures/cp_app/ensemble_models_smallML_120_100/300',
    '4-MLFeatures/cp_app/ensemble_models_smallML_120_100/350',
    '4-MLFeatures/cp_app/ensemble_models_smallML_120_100/400',
    '4-MLFeatures/models'
]
local_directories = [
    '/models/cp_app/ensemble_models_smallML_120_100/300',
    '/models/cp_app/ensemble_models_smallML_120_100/350',
    '/models/cp_app/ensemble_models_smallML_120_100/400',
    '/models/stability'
]

path_map = dict(zip(github_paths, local_directories))

for github_path in github_paths:
    files_info = get_files_from_github(repo, github_path)
    for file_info in files_info:
        if file_info['type'] == 'file':
            raw_url = file_info['download_url']
            local_path_suffix = path_map[github_path].lstrip('/')
            local_file_path = os.path.join(package_directory, local_path_suffix, os.path.basename(file_info['path']))
            download_file(raw_url, local_file_path)

from CoREMOF.models.cp_app.descriptors import cv_features
from CoREMOF.models.cp_app.featurizer import featurize_structure
from CoREMOF.models.cp_app.predictions import predict_Cv_ensemble_structure_multitemperatures


def pacman(structure, output_folder="result_pacman", charge_type="DDEC6", digits=10, atom_type=True, neutral=True, keep_connect=False):

    results_eb = {}
    
    if not os.path.exists(output_folder):
        os.makedirs(output_folder, exist_ok=True)
    
    name = os.path.basename(structure).replace(".cif", "")
    results_eb["Name"] = name

    try:
        pmcharge.predict(
            cif_file=structure,
            charge_type=charge_type,
            digits=digits,
            atom_type=atom_type,
            neutral=neutral,
            keep_connect=keep_connect
        )
        
        pbe, bandgap = pmcharge.Energy(cif_file=structure)
        
        results_eb["PBE Energy"] = pbe
        results_eb["Bandgap"] = bandgap
      
        shutil.move(structure.replace(".cif","_pacman.cif"),
                    os.path.join(output_folder,
                                structure.split("/")[-1].replace(".cif","_pacman.cif")))

        return results_eb
    
    except Exception as e:
        print(e)

def cp(structure, T=[300, 350, 400]):
    
    name = os.path.basename(structure).replace(".cif", "")
    featurize_structure(structure, verbos=False, saveto="features.csv")
    
    predict_Cv_ensemble_structure_multitemperatures(
                                                    path_to_models=package_directory+"models/cp_app/ensemble_models_smallML_120_100",
                                                    structure_name=name + ".cif",
                                                    features_file="features.csv", 
                                                    FEATURES=cv_features,
                                                    temperatures=T,
                                                    save_to="cp.csv"
                                                    )
    result_ = pd.read_csv("cp.csv")
    result_cp = {}
    result_cp["unit"] = "J/g/K", "J/mol/K"

    for t in T:
        result_cp[str(t)+"_mean"] = [result_["Cv_gravimetric_"+str(t)+"_mean"].iloc[0],
                                        result_["Cv_molar_"+str(t)+"_mean"].iloc[0]]
        result_cp[str(t)+"_std"] = [result_["Cv_gravimetric_"+str(t)+"_std"].iloc[0],
                                        result_["Cv_molar_"+str(t)+"_std"].iloc[0]]

    os.remove("features.csv")
    os.remove("cp.csv")

    return result_cp


def precision(y_true, y_pred):
    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    predicted_positives = K.sum(K.round(K.clip(y_pred, 0, 1)))
    precision = true_positives / (predicted_positives + K.epsilon())
    return precision
def recall(y_true, y_pred):
    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    possible_positives = K.sum(K.round(K.clip(y_true, 0, 1)))
    recall = true_positives / (possible_positives + K.epsilon())
    return recall
def f1(y_true, y_pred):
    p = precision(y_true, y_pred)
    r = recall(y_true, y_pred)
    return 2 * ((p * r) / (p + r + K.epsilon()))


import tensorflow as tf
# tf.config.set_visible_devices([], 'GPU')

# os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
def stability(structure):
   
    solvent_feature_names = [
                            'f-chi-0-all', 'f-chi-1-all', 'f-chi-2-all', 'f-chi-3-all', 'f-Z-0-all',
                            'f-Z-1-all', 'f-Z-2-all', 'f-Z-3-all', 'f-I-0-all', 'f-I-1-all', 'f-I-2-all',
                            'f-I-3-all', 'f-T-0-all', 'f-T-1-all', 'f-T-2-all', 'f-T-3-all', 'f-S-0-all',
                            'f-S-1-all', 'f-S-2-all', 'f-S-3-all', 'mc-chi-0-all', 'mc-chi-1-all',
                            'mc-chi-2-all', 'mc-chi-3-all', 'mc-Z-0-all', 'mc-Z-1-all', 'mc-Z-2-all',
                            'mc-Z-3-all', 'mc-I-1-all', 'mc-I-2-all', 'mc-I-3-all', 'mc-T-0-all',
                            'mc-T-1-all', 'mc-T-2-all', 'mc-T-3-all', 'mc-S-0-all', 'mc-S-1-all',
                            'mc-S-2-all', 'mc-S-3-all', 'D_mc-chi-1-all', 'D_mc-chi-2-all',
                            'D_mc-chi-3-all', 'D_mc-Z-1-all', 'D_mc-Z-2-all', 'D_mc-Z-3-all',
                            'D_mc-T-1-all', 'D_mc-T-2-all', 'D_mc-T-3-all', 'D_mc-S-1-all',
                            'D_mc-S-2-all', 'D_mc-S-3-all', 'f-lig-chi-0', 'f-lig-chi-1',
                            'f-lig-chi-2', 'f-lig-chi-3', 'f-lig-Z-0', 'f-lig-Z-1', 'f-lig-Z-2',
                            'f-lig-Z-3', 'f-lig-I-0', 'f-lig-I-1', 'f-lig-I-2', 'f-lig-I-3',
                            'f-lig-T-0', 'f-lig-T-1', 'f-lig-T-2', 'f-lig-T-3', 'f-lig-S-0',
                            'f-lig-S-1', 'f-lig-S-2', 'f-lig-S-3', 'lc-chi-0-all', 'lc-chi-1-all',
                            'lc-chi-2-all', 'lc-chi-3-all', 'lc-Z-0-all', 'lc-Z-1-all', 'lc-Z-2-all',
                            'lc-Z-3-all', 'lc-I-1-all', 'lc-I-2-all', 'lc-I-3-all', 'lc-T-0-all',
                            'lc-T-1-all', 'lc-T-2-all', 'lc-T-3-all', 'lc-S-0-all', 'lc-S-1-all',
                            'lc-S-2-all', 'lc-S-3-all', 'D_lc-chi-1-all', 'D_lc-chi-2-all',
                            'D_lc-chi-3-all', 'D_lc-Z-1-all', 'D_lc-Z-2-all', 'D_lc-Z-3-all',
                            'D_lc-T-1-all', 'D_lc-T-2-all', 'D_lc-T-3-all', 'D_lc-S-1-all',
                            'D_lc-S-2-all', 'D_lc-S-3-all', 'func-chi-0-all', 'func-chi-1-all',
                            'func-chi-2-all', 'func-chi-3-all', 'func-Z-0-all', 'func-Z-1-all',
                            'func-Z-2-all', 'func-Z-3-all', 'func-I-0-all', 'func-I-1-all',
                            'func-I-2-all', 'func-I-3-all', 'func-T-0-all', 'func-T-1-all',
                            'func-T-2-all', 'func-T-3-all', 'func-S-0-all', 'func-S-1-all',
                            'func-S-2-all', 'func-S-3-all', 'D_func-chi-1-all', 'D_func-chi-2-all',
                            'D_func-chi-3-all', 'D_func-Z-1-all', 'D_func-Z-2-all', 'D_func-Z-3-all',
                            'D_func-T-1-all', 'D_func-T-2-all', 'D_func-T-3-all', 'D_func-S-1-all',
                            'D_func-S-2-all', 'D_func-S-3-all']

    thermal_feature_names = ['f-chi-0-all', 'f-chi-1-all', 'f-chi-2-all', 'f-chi-3-all',
                            'f-Z-0-all', 'f-Z-1-all', 'f-Z-2-all', 'f-Z-3-all', 'f-I-0-all',
                            'f-I-1-all', 'f-I-2-all', 'f-I-3-all', 'f-T-0-all', 'f-T-1-all',
                            'f-T-2-all', 'f-T-3-all', 'f-S-0-all', 'f-S-1-all', 'f-S-2-all',
                            'f-S-3-all', 'mc-chi-0-all', 'mc-chi-1-all', 'mc-chi-2-all',
                            'mc-chi-3-all', 'mc-Z-0-all', 'mc-Z-1-all', 'mc-Z-2-all',
                            'mc-Z-3-all', 'mc-I-1-all', 'mc-I-2-all', 'mc-I-3-all',
                            'mc-T-0-all', 'mc-T-1-all', 'mc-T-2-all', 'mc-T-3-all',
                            'mc-S-0-all', 'mc-S-1-all', 'mc-S-2-all', 'mc-S-3-all',
                            'D_mc-chi-1-all', 'D_mc-chi-2-all', 'D_mc-chi-3-all',
                            'D_mc-Z-1-all', 'D_mc-Z-2-all', 'D_mc-Z-3-all', 'D_mc-T-1-all',
                            'D_mc-T-2-all', 'D_mc-T-3-all', 'D_mc-S-1-all', 'D_mc-S-2-all',
                            'D_mc-S-3-all', 'f-lig-chi-0', 'f-lig-chi-1', 'f-lig-chi-2',
                            'f-lig-chi-3', 'f-lig-Z-0', 'f-lig-Z-1', 'f-lig-Z-2', 'f-lig-Z-3',
                            'f-lig-I-0', 'f-lig-I-1', 'f-lig-I-2', 'f-lig-I-3', 'f-lig-T-0',
                            'f-lig-T-1', 'f-lig-T-2', 'f-lig-T-3', 'f-lig-S-0', 'f-lig-S-1',
                            'f-lig-S-2', 'f-lig-S-3', 'lc-chi-0-all', 'lc-chi-1-all', 'lc-chi-2-all',
                            'lc-chi-3-all', 'lc-Z-0-all', 'lc-Z-1-all', 'lc-Z-2-all', 'lc-Z-3-all',
                            'lc-I-1-all', 'lc-I-2-all', 'lc-I-3-all', 'lc-T-0-all', 'lc-T-1-all',
                            'lc-T-2-all', 'lc-T-3-all', 'lc-S-0-all', 'lc-S-1-all', 'lc-S-2-all',
                            'lc-S-3-all', 'D_lc-chi-1-all', 'D_lc-chi-2-all', 'D_lc-chi-3-all',
                            'D_lc-Z-1-all', 'D_lc-Z-2-all', 'D_lc-Z-3-all', 'D_lc-T-1-all',
                            'D_lc-T-2-all', 'D_lc-T-3-all', 'D_lc-S-1-all', 'D_lc-S-2-all',
                            'D_lc-S-3-all', 'func-chi-0-all', 'func-chi-1-all', 'func-chi-2-all',
                            'func-chi-3-all', 'func-Z-0-all', 'func-Z-1-all', 'func-Z-2-all',
                            'func-Z-3-all', 'func-I-0-all', 'func-I-1-all', 'func-I-2-all',
                            'func-I-3-all', 'func-T-0-all', 'func-T-1-all', 'func-T-2-all',
                            'func-T-3-all', 'func-S-0-all', 'func-S-1-all', 'func-S-2-all',
                            'func-S-3-all', 'D_func-chi-1-all', 'D_func-chi-2-all', 'D_func-chi-3-all',
                            'D_func-Z-1-all', 'D_func-Z-2-all', 'D_func-Z-3-all', 'D_func-T-1-all',
                            'D_func-T-2-all', 'D_func-T-3-all', 'D_func-S-1-all', 'D_func-S-2-all',
                            'D_func-S-3-all']

    water_feature_names = ['mc-Z-3-all', 'D_mc-Z-3-all', 'D_mc-Z-2-all',
                            'D_mc-Z-1-all', 'mc-chi-3-all', 'mc-Z-1-all',
                            'mc-Z-0-all', 'D_mc-chi-2-all', 'f-lig-Z-2',
                            "ASA",
                            'f-lig-I-0', 'func-S-1-all']

    result_stability = {}
    result_stability["unit"] = "nan, °C, nan"

    dependencies = {'precision':precision,'recall':recall,'f1':f1}
    solvent_model = keras.models.load_model(package_directory+'/models/stability/final_model_flag_few_epochs.h5', custom_objects=dependencies)
    thermal_model = keras.models.load_model(package_directory+'/models/stability/final_model_T_few_epochs.h5', custom_objects=dependencies)
    with open(package_directory+'/models/stability/solvent_scaler.pkl', 'rb') as f:
        solvent_scaler = pkl.load(f)
    with open(package_directory+'/models/stability/thermal_x_scaler.pkl', 'rb') as f:
        thermal_x_scaler = pkl.load(f)
    with open(package_directory+'/models/stability/thermal_y_scaler.pkl', 'rb') as f:
        thermal_y_scaler = pkl.load(f)
    with open(package_directory+'/models/stability/water_model.pkl', 'rb') as f:
        water_model = cloudpickle.load(f)
    with open(package_directory+'/models/stability/water_scaler.pkl', 'rb') as f:
        water_scaler = pkl.load(f)

    results_pd = Zeopp.PoreDiameter(structure)
    results_pv = Zeopp.PoreVolume(structure, 1.86, 1.86, 10000, True)
    results_sa = Zeopp.SurfaceArea(structure, 1.86, 1.86, 10000, True)
    result_v = Volume(structure)
    results_sa_1_4 = Zeopp.SurfaceArea(structure, 1.4, 1.4, 10000, True)
    result_RACs = RACs(structure)
    
    '''
        zeopp4sol in the start
        LCD,
        PLD,
        LFPD,
        GPOAV_1_86,
        GPONAV_1_86,
        GPOV_1_86,
        GSA_1_86,
        POAV_1_86,
        POAV_vol_frac_1_86,
        PONAV_1_86,
        PONAV_vol_frac_1_86,
        VPOV_1_86,
        VSA_1_86,
        cell_volume
    '''

    X_solvent = [results_pd["LCD"],
                    results_pd["PLD"],
                    results_pd["LFPD"],
                    results_pv["PV"][1],
                    results_pv["NPV"][1],
                    results_pv["PV"][1] + results_pv["NPV"][1],
                    results_sa["ASA"][2],
                    results_pv["PV"][0],
                    results_pv["VF"],
                    results_pv["NPV"][0],
                    results_pv["NVF"],
                    results_pv["VF"] + results_pv["NVF"],
                    results_sa["ASA"][1],
                    result_v["total_volume"]
                ]
    
    for fn_sol in solvent_feature_names:
        try:
            X_solvent.append(result_RACs["Metal"][fn_sol])
        except:
            try:
                X_solvent.append(result_RACs["Linker"][fn_sol])
            except:
                X_solvent.append(result_RACs["Function-group"][fn_sol])

    # X_solvent = np.array(X_solvent).reshape(-1, 1).flatten().reshape(1, -1).tolist()
    
    X_solvent = solvent_scaler.transform([X_solvent])
    solvent_model_prob = solvent_model.predict(X_solvent)
    solvent_model_prob = solvent_model_prob.flatten()
    result_stability["solvent removal probability"] = float(solvent_model_prob[0])

    '''
        zeopp4therm in the end
        LCD,
        PLD,
        LFPD,
        GPOAV_1_86,
        GPONAV_1_86,
        GPOV_1_86,
        GSA_1_86,
        POAV_1_86,
        POAV_vol_frac_1_86,
        PONAV_1_86,
        PONAV_vol_frac_1_86,
        VPOV_1_86,
        VSA_1_86,
        cell_volume
    '''

    X_thermal = []
    for fn_them in thermal_feature_names:
        try:
            X_thermal.append(result_RACs["Metal"][fn_them])
        except:
            try:
                X_thermal.append(result_RACs["Linker"][fn_them])
            except:
                X_thermal.append(result_RACs["Function-group"][fn_them])

    for zeo_them in [results_pd["LCD"],
                    results_pd["PLD"],
                    results_pd["LFPD"],
                    results_pv["PV"][1],
                    results_pv["NPV"][1],
                    results_pv["PV"][1] + results_pv["NPV"][1],
                    results_sa["ASA"][2],
                    results_pv["PV"][0],
                    results_pv["VF"],
                    results_pv["NPV"][0],
                    results_pv["NVF"],
                    results_pv["VF"] + results_pv["NVF"],
                    results_sa["ASA"][1],
                    result_v["total_volume"]
                ]:
        
        X_thermal.append(zeo_them)
    X_thermal = thermal_x_scaler.transform([X_thermal])
    thermal_model_pred = thermal_y_scaler.inverse_transform(thermal_model.predict(X_thermal))
    thermal_model_pred = np.round(thermal_model_pred, 1)
    thermal_model_pred = thermal_model_pred.flatten()
    result_stability["thermal stability"] = float(thermal_model_pred[0])

    X_water = []
    for fn_water in water_feature_names:
        try:
            X_water.append(result_RACs["Metal"][fn_water])
        except:
            try:
                X_water.append(result_RACs["Linker"][fn_water])
            except:
                try:
                    X_water.append(result_RACs["Function-group"][fn_water])
                except:
                    X_water.append(results_sa_1_4[fn_water][2])

    X_water = water_scaler.transform([X_water])

    water_model_prob = water_model.predict_proba(X_water)[:,1]
    # water_model_label = water_model.predict(X_water)
    result_stability["water probability"] = float(water_model_prob[0])

    return result_stability